"""
Indexing support for pd.DataFrame type.
"""
import operator
import numpy as np
import pandas as pd
from numba.core import cgutils, types
from numba.core.typing.templates import AbstractTemplate, infer_global
from numba.extending import intrinsic, lower_builtin, lower_cast, make_attribute_wrapper, models, overload, overload_attribute, register_model
import bodo
from bodo.hiframes.pd_dataframe_ext import DataFrameType, check_runtime_cols_unsupported
from bodo.utils.transform import gen_const_tup
from bodo.utils.typing import BodoError, get_overload_const_int, get_overload_const_list, get_overload_const_str, is_immutable_array, is_list_like_index_type, is_overload_constant_int, is_overload_constant_list, is_overload_constant_str, raise_bodo_error


@infer_global(operator.getitem)
class DataFrameGetItemTemplate(AbstractTemplate):

    def generic(self, args, kws):
        assert not kws
        assert len(args) == 2
        check_runtime_cols_unsupported(args[0], 'DataFrame getitem (df[])')
        if isinstance(args[0], DataFrameType):
            return self.typecheck_df_getitem(args)
        elif isinstance(args[0], DataFrameLocType):
            return self.typecheck_loc_getitem(args)
        else:
            return

    def typecheck_loc_getitem(self, args):
        I = args[0]
        idx = args[1]
        df = I.df_type
        if isinstance(df.columns[0], tuple):
            raise_bodo_error(
                'DataFrame.loc[] getitem (location-based indexing) with multi-indexed columns not supported yet'
                )
        if is_list_like_index_type(idx) and idx.dtype == types.bool_:
            sxwi__dvo = idx
            fvcka__aej = df.data
            ktf__qxi = df.columns
            nei__sxwa = self.replace_range_with_numeric_idx_if_needed(df,
                sxwi__dvo)
            mxfp__imez = DataFrameType(fvcka__aej, nei__sxwa, ktf__qxi)
            return mxfp__imez(*args)
        if isinstance(idx, types.BaseTuple) and len(idx) == 2:
            xpin__hrgei = idx.types[0]
            rku__nlw = idx.types[1]
            if isinstance(xpin__hrgei, types.Integer):
                if not isinstance(df.index, bodo.hiframes.pd_index_ext.
                    RangeIndexType):
                    raise_bodo_error(
                        'Dataframe.loc[int, col_ind] getitem only supported for dataframes with RangeIndexes'
                        )
                if is_overload_constant_str(rku__nlw):
                    losw__wwpqm = get_overload_const_str(rku__nlw)
                    if losw__wwpqm not in df.columns:
                        raise_bodo_error(
                            'dataframe {} does not include column {}'.
                            format(df, losw__wwpqm))
                    nsg__onms = df.columns.index(losw__wwpqm)
                    return df.data[nsg__onms].dtype(*args)
                if isinstance(rku__nlw, types.UnicodeType):
                    raise_bodo_error(
                        f'DataFrame.loc[] getitem (location-based indexing) requires constant column names. For more information, see https://docs.bodo.ai/latest/source/programming_with_bodo/require_constants.html.'
                        )
                else:
                    raise_bodo_error(
                        f'DataFrame.loc[] getitem (location-based indexing) using {idx} not supported yet.'
                        )
            if is_list_like_index_type(xpin__hrgei
                ) and xpin__hrgei.dtype == types.bool_ or isinstance(
                xpin__hrgei, types.SliceType):
                nei__sxwa = self.replace_range_with_numeric_idx_if_needed(df,
                    xpin__hrgei)
                if is_overload_constant_str(rku__nlw):
                    txyi__jtj = get_overload_const_str(rku__nlw)
                    if txyi__jtj not in df.columns:
                        raise_bodo_error(
                            f'dataframe {df} does not include column {txyi__jtj}'
                            )
                    nsg__onms = df.columns.index(txyi__jtj)
                    wnyxk__byt = df.data[nsg__onms]
                    mjymi__zblf = wnyxk__byt.dtype
                    xgnqp__lrc = types.literal(df.columns[nsg__onms])
                    mxfp__imez = bodo.SeriesType(mjymi__zblf, wnyxk__byt,
                        nei__sxwa, xgnqp__lrc)
                    return mxfp__imez(*args)
                if isinstance(rku__nlw, types.UnicodeType):
                    raise_bodo_error(
                        f'DataFrame.loc[] getitem (location-based indexing) requires constant column names. For more information, see https://docs.bodo.ai/latest/source/programming_with_bodo/require_constants.html.'
                        )
                elif is_overload_constant_list(rku__nlw):
                    nxl__xwugy = get_overload_const_list(rku__nlw)
                    zli__mehh = types.unliteral(rku__nlw)
                    if zli__mehh.dtype == types.bool_:
                        if len(df.columns) != len(nxl__xwugy):
                            raise_bodo_error(
                                f'dataframe {df} has {len(df.columns)} columns, but boolean array used with DataFrame.loc[] {nxl__xwugy} has {len(nxl__xwugy)} values'
                                )
                        lcy__iub = []
                        cxpb__jyhmv = []
                        for wfn__ipdvu in range(len(nxl__xwugy)):
                            if nxl__xwugy[wfn__ipdvu]:
                                lcy__iub.append(df.columns[wfn__ipdvu])
                                cxpb__jyhmv.append(df.data[wfn__ipdvu])
                        ruo__qej = tuple()
                        mxfp__imez = DataFrameType(tuple(cxpb__jyhmv),
                            nei__sxwa, tuple(lcy__iub))
                        return mxfp__imez(*args)
                    elif zli__mehh.dtype == bodo.string_type:
                        ruo__qej, cxpb__jyhmv = self.get_kept_cols_and_data(df,
                            nxl__xwugy)
                        mxfp__imez = DataFrameType(cxpb__jyhmv, nei__sxwa,
                            ruo__qej)
                        return mxfp__imez(*args)
        raise_bodo_error(
            f'DataFrame.loc[] getitem (location-based indexing) using {idx} not supported yet. If you are trying to select a subset of the columns by passing a list of column names, that list must be a compile time constant. See https://docs.bodo.ai/latest/source/programming_with_bodo/require_constants.html.'
            )

    def typecheck_df_getitem(self, args):
        df = args[0]
        ind = args[1]
        if is_overload_constant_str(ind) or is_overload_constant_int(ind):
            ind_val = get_overload_const_str(ind) if is_overload_constant_str(
                ind) else get_overload_const_int(ind)
            if isinstance(df.columns[0], tuple):
                lcy__iub = []
                cxpb__jyhmv = []
                for wfn__ipdvu, hkjik__sca in enumerate(df.columns):
                    if hkjik__sca[0] != ind_val:
                        continue
                    lcy__iub.append(hkjik__sca[1] if len(hkjik__sca) == 2 else
                        hkjik__sca[1:])
                    cxpb__jyhmv.append(df.data[wfn__ipdvu])
                wnyxk__byt = tuple(cxpb__jyhmv)
                piie__xqmf = df.index
                ubyhv__mjd = tuple(lcy__iub)
                mxfp__imez = DataFrameType(wnyxk__byt, piie__xqmf, ubyhv__mjd)
                return mxfp__imez(*args)
            else:
                if ind_val not in df.columns:
                    raise_bodo_error('dataframe {} does not include column {}'
                        .format(df, ind_val))
                nsg__onms = df.columns.index(ind_val)
                wnyxk__byt = df.data[nsg__onms]
                mjymi__zblf = wnyxk__byt.dtype
                piie__xqmf = df.index
                xgnqp__lrc = types.literal(df.columns[nsg__onms])
                mxfp__imez = bodo.SeriesType(mjymi__zblf, wnyxk__byt,
                    piie__xqmf, xgnqp__lrc)
                return mxfp__imez(*args)
        if isinstance(ind, types.Integer) or isinstance(ind, types.UnicodeType
            ):
            raise_bodo_error(
                'df[] getitem selecting a subset of columns requires providing constant column names. For more information, see https://docs.bodo.ai/latest/source/programming_with_bodo/require_constants.html'
                )
        if is_list_like_index_type(ind
            ) and ind.dtype == types.bool_ or isinstance(ind, types.SliceType):
            wnyxk__byt = df.data
            piie__xqmf = self.replace_range_with_numeric_idx_if_needed(df, ind)
            ubyhv__mjd = df.columns
            mxfp__imez = DataFrameType(wnyxk__byt, piie__xqmf, ubyhv__mjd,
                is_table_format=df.is_table_format)
            return mxfp__imez(*args)
        elif is_overload_constant_list(ind):
            tvdto__yaen = get_overload_const_list(ind)
            ubyhv__mjd, wnyxk__byt = self.get_kept_cols_and_data(df,
                tvdto__yaen)
            piie__xqmf = df.index
            mxfp__imez = DataFrameType(wnyxk__byt, piie__xqmf, ubyhv__mjd)
            return mxfp__imez(*args)
        raise_bodo_error(
            f'df[] getitem using {ind} not supported. If you are trying to select a subset of the columns, you must provide the column names you are selecting as a constant. See https://docs.bodo.ai/latest/source/programming_with_bodo/require_constants.html'
            )

    def get_kept_cols_and_data(self, df, cols_to_keep_list):
        for pzsex__zyy in cols_to_keep_list:
            if pzsex__zyy not in df.columns:
                raise_bodo_error('Column {} not found in dataframe columns {}'
                    .format(pzsex__zyy, df.columns))
        ubyhv__mjd = tuple(cols_to_keep_list)
        wnyxk__byt = tuple(df.data[df.columns.index(xxv__khoxk)] for
            xxv__khoxk in ubyhv__mjd)
        return ubyhv__mjd, wnyxk__byt

    def replace_range_with_numeric_idx_if_needed(self, df, ind):
        nei__sxwa = bodo.hiframes.pd_index_ext.NumericIndexType(types.int64,
            df.index.name_typ) if not isinstance(ind, types.SliceType
            ) and isinstance(df.index, bodo.hiframes.pd_index_ext.
            RangeIndexType) else df.index
        return nei__sxwa


DataFrameGetItemTemplate._no_unliteral = True


@lower_builtin(operator.getitem, DataFrameType, types.Any)
def getitem_df_lower(context, builder, sig, args):
    impl = df_getitem_overload(*sig.args)
    return context.compile_internal(builder, impl, sig, args)


def df_getitem_overload(df, ind):
    if not isinstance(df, DataFrameType):
        return
    if is_overload_constant_str(ind) or is_overload_constant_int(ind):
        ind_val = get_overload_const_str(ind) if is_overload_constant_str(ind
            ) else get_overload_const_int(ind)
        if isinstance(df.columns[0], tuple):
            lcy__iub = []
            cxpb__jyhmv = []
            for wfn__ipdvu, hkjik__sca in enumerate(df.columns):
                if hkjik__sca[0] != ind_val:
                    continue
                lcy__iub.append(hkjik__sca[1] if len(hkjik__sca) == 2 else
                    hkjik__sca[1:])
                cxpb__jyhmv.append(
                    'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {})'
                    .format(wfn__ipdvu))
            llc__qshk = 'def impl(df, ind):\n'
            jqe__vul = 'bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)'
            return bodo.hiframes.dataframe_impl._gen_init_df(llc__qshk,
                lcy__iub, ', '.join(cxpb__jyhmv), jqe__vul)
        if ind_val not in df.columns:
            raise_bodo_error('dataframe {} does not include column {}'.
                format(df, ind_val))
        col_no = df.columns.index(ind_val)
        return lambda df, ind: bodo.hiframes.pd_series_ext.init_series(bodo
            .hiframes.pd_dataframe_ext.get_dataframe_data(df, col_no), bodo
            .hiframes.pd_dataframe_ext.get_dataframe_index(df), ind_val)
    if is_overload_constant_list(ind):
        tvdto__yaen = get_overload_const_list(ind)
        for pzsex__zyy in tvdto__yaen:
            if pzsex__zyy not in df.columns:
                raise_bodo_error('Column {} not found in dataframe columns {}'
                    .format(pzsex__zyy, df.columns))
        cxpb__jyhmv = ', '.join(
            'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {}).copy()'
            .format(df.columns.index(pzsex__zyy)) for pzsex__zyy in tvdto__yaen
            )
        llc__qshk = 'def impl(df, ind):\n'
        jqe__vul = 'bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)'
        return bodo.hiframes.dataframe_impl._gen_init_df(llc__qshk,
            tvdto__yaen, cxpb__jyhmv, jqe__vul)
    if is_list_like_index_type(ind) and ind.dtype == types.bool_ or isinstance(
        ind, types.SliceType):
        llc__qshk = 'def impl(df, ind):\n'
        if not isinstance(ind, types.SliceType):
            llc__qshk += (
                '  ind = bodo.utils.conversion.coerce_to_ndarray(ind)\n')
        jqe__vul = (
            'bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)[ind]')
        if df.is_table_format:
            cxpb__jyhmv = (
                f'bodo.hiframes.pd_dataframe_ext.get_dataframe_table(df)[ind]')
        else:
            cxpb__jyhmv = ', '.join(
                f'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {df.columns.index(pzsex__zyy)})[ind]'
                 for pzsex__zyy in df.columns)
        return bodo.hiframes.dataframe_impl._gen_init_df(llc__qshk, df.
            columns, cxpb__jyhmv, jqe__vul, out_df_type=df)
    raise_bodo_error('df[] getitem using {} not supported'.format(ind))


@overload(operator.setitem, no_unliteral=True)
def df_setitem_overload(df, idx, val):
    check_runtime_cols_unsupported(df, 'DataFrame setitem (df[])')
    if not isinstance(df, DataFrameType):
        return
    raise_bodo_error('DataFrame setitem: transform necessary')


class DataFrameILocType(types.Type):

    def __init__(self, df_type):
        self.df_type = df_type
        xxv__khoxk = 'DataFrameILocType({})'.format(df_type)
        super(DataFrameILocType, self).__init__(xxv__khoxk)

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)
    ndim = 2


@register_model(DataFrameILocType)
class DataFrameILocModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        wuh__rchec = [('obj', fe_type.df_type)]
        super(DataFrameILocModel, self).__init__(dmm, fe_type, wuh__rchec)


make_attribute_wrapper(DataFrameILocType, 'obj', '_obj')


@intrinsic
def init_dataframe_iloc(typingctx, obj=None):

    def codegen(context, builder, signature, args):
        fatsf__ctu, = args
        uys__jcde = signature.return_type
        dkj__rfdk = cgutils.create_struct_proxy(uys__jcde)(context, builder)
        dkj__rfdk.obj = fatsf__ctu
        context.nrt.incref(builder, signature.args[0], fatsf__ctu)
        return dkj__rfdk._getvalue()
    return DataFrameILocType(obj)(obj), codegen


@overload_attribute(DataFrameType, 'iloc')
def overload_dataframe_iloc(df):
    check_runtime_cols_unsupported(df, 'DataFrame.iloc')
    return lambda df: bodo.hiframes.dataframe_indexing.init_dataframe_iloc(df)


@overload(operator.getitem, no_unliteral=True)
def overload_iloc_getitem(I, idx):
    if not isinstance(I, DataFrameILocType):
        return
    df = I.df_type
    if isinstance(idx, types.Integer):
        return _gen_iloc_getitem_row_impl(df, df.columns, 'idx')
    if isinstance(idx, types.BaseTuple) and len(idx) == 2 and not isinstance(
        idx[1], types.SliceType):
        if not (is_overload_constant_list(idx.types[1]) or
            is_overload_constant_int(idx.types[1])):
            raise_bodo_error(
                'idx2 in df.iloc[idx1, idx2] should be a constant integer or constant list of integers. For more information, see https://docs.bodo.ai/latest/source/programming_with_bodo/require_constants.html.'
                )
        tcenf__ioqxt = len(df.data)
        if is_overload_constant_int(idx.types[1]):
            is_out_series = True
            rwsi__dit = get_overload_const_int(idx.types[1])
            if rwsi__dit < 0 or rwsi__dit >= tcenf__ioqxt:
                raise BodoError(
                    'df.iloc: column integer must refer to a valid column number'
                    )
            kxrh__hypi = [rwsi__dit]
        else:
            is_out_series = False
            kxrh__hypi = get_overload_const_list(idx.types[1])
            if any(not isinstance(ind, int) or ind < 0 or ind >=
                tcenf__ioqxt for ind in kxrh__hypi):
                raise BodoError(
                    'df.iloc: column list must be integers referring to a valid column number'
                    )
        col_names = tuple(pd.Series(df.columns, dtype=object)[kxrh__hypi])
        if isinstance(idx.types[0], types.Integer):
            if isinstance(idx.types[1], types.Integer):
                rwsi__dit = kxrh__hypi[0]

                def impl(I, idx):
                    df = I._obj
                    return bodo.utils.conversion.box_if_dt64(bodo.hiframes.
                        pd_dataframe_ext.get_dataframe_data(df, rwsi__dit)[
                        idx[0]])
                return impl
            return _gen_iloc_getitem_row_impl(df, col_names, 'idx[0]')
        if is_list_like_index_type(idx.types[0]) and isinstance(idx.types[0
            ].dtype, (types.Integer, types.Boolean)) or isinstance(idx.
            types[0], types.SliceType):
            return _gen_iloc_getitem_bool_slice_impl(df, col_names, idx.
                types[0], 'idx[0]', is_out_series)
    if is_list_like_index_type(idx) and isinstance(idx.dtype, (types.
        Integer, types.Boolean)) or isinstance(idx, types.SliceType):
        return _gen_iloc_getitem_bool_slice_impl(df, df.columns, idx, 'idx',
            False)
    if isinstance(idx, types.BaseTuple) and len(idx) == 2 and isinstance(idx
        [0], types.SliceType) and isinstance(idx[1], types.SliceType):
        raise_bodo_error(
            'slice2 in df.iloc[slice1,slice2] should be constant. For more information, see https://docs.bodo.ai/latest/source/programming_with_bodo/require_constants.html.'
            )
    raise_bodo_error(f'df.iloc[] getitem using {idx} not supported')


def _gen_iloc_getitem_bool_slice_impl(df, col_names, idx_typ, idx,
    is_out_series):
    llc__qshk = 'def impl(I, idx):\n'
    llc__qshk += '  df = I._obj\n'
    if isinstance(idx_typ, types.SliceType):
        llc__qshk += f'  idx_t = {idx}\n'
    else:
        llc__qshk += (
            f'  idx_t = bodo.utils.conversion.coerce_to_ndarray({idx})\n')
    jqe__vul = 'bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)[idx_t]'
    cxpb__jyhmv = ', '.join(
        f'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {df.columns.index(pzsex__zyy)})[idx_t]'
         for pzsex__zyy in col_names)
    if is_out_series:
        jdcj__wlkep = f"'{col_names[0]}'" if isinstance(col_names[0], str
            ) else f'{col_names[0]}'
        llc__qshk += f"""  return bodo.hiframes.pd_series_ext.init_series({cxpb__jyhmv}, {jqe__vul}, {jdcj__wlkep})
"""
        avlp__lltk = {}
        exec(llc__qshk, {'bodo': bodo}, avlp__lltk)
        return avlp__lltk['impl']
    return bodo.hiframes.dataframe_impl._gen_init_df(llc__qshk, col_names,
        cxpb__jyhmv, jqe__vul)


def _gen_iloc_getitem_row_impl(df, col_names, idx):
    llc__qshk = 'def impl(I, idx):\n'
    llc__qshk += '  df = I._obj\n'
    kblkt__zxjq = ', '.join(
        f'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {df.columns.index(pzsex__zyy)})[{idx}]'
         for pzsex__zyy in col_names)
    llc__qshk += f"""  row_idx = bodo.hiframes.pd_index_ext.init_heter_index({gen_const_tup(col_names)}, None)
"""
    llc__qshk += f"""  return bodo.hiframes.pd_series_ext.init_series(({kblkt__zxjq},), row_idx, None)
"""
    avlp__lltk = {}
    exec(llc__qshk, {'bodo': bodo}, avlp__lltk)
    impl = avlp__lltk['impl']
    return impl


@overload(operator.setitem, no_unliteral=True)
def df_iloc_setitem_overload(df, idx, val):
    if not isinstance(df, DataFrameILocType):
        return
    raise_bodo_error(
        f'DataFrame.iloc setitem unsupported for dataframe {df.df_type}, index {idx}, value {val}'
        )


class DataFrameLocType(types.Type):

    def __init__(self, df_type):
        self.df_type = df_type
        xxv__khoxk = 'DataFrameLocType({})'.format(df_type)
        super(DataFrameLocType, self).__init__(xxv__khoxk)

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)
    ndim = 2


@register_model(DataFrameLocType)
class DataFrameLocModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        wuh__rchec = [('obj', fe_type.df_type)]
        super(DataFrameLocModel, self).__init__(dmm, fe_type, wuh__rchec)


make_attribute_wrapper(DataFrameLocType, 'obj', '_obj')


@intrinsic
def init_dataframe_loc(typingctx, obj=None):

    def codegen(context, builder, signature, args):
        fatsf__ctu, = args
        kkyfl__bfrt = signature.return_type
        lhv__luxh = cgutils.create_struct_proxy(kkyfl__bfrt)(context, builder)
        lhv__luxh.obj = fatsf__ctu
        context.nrt.incref(builder, signature.args[0], fatsf__ctu)
        return lhv__luxh._getvalue()
    return DataFrameLocType(obj)(obj), codegen


@overload_attribute(DataFrameType, 'loc')
def overload_dataframe_loc(df):
    check_runtime_cols_unsupported(df, 'DataFrame.loc')
    return lambda df: bodo.hiframes.dataframe_indexing.init_dataframe_loc(df)


@lower_builtin(operator.getitem, DataFrameLocType, types.Any)
def loc_getitem_lower(context, builder, sig, args):
    impl = overload_loc_getitem(*sig.args)
    return context.compile_internal(builder, impl, sig, args)


def overload_loc_getitem(I, idx):
    if not isinstance(I, DataFrameLocType):
        return
    df = I.df_type
    if is_list_like_index_type(idx) and idx.dtype == types.bool_:
        llc__qshk = 'def impl(I, idx):\n'
        llc__qshk += '  df = I._obj\n'
        llc__qshk += '  idx_t = bodo.utils.conversion.coerce_to_ndarray(idx)\n'
        jqe__vul = (
            'bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)[idx_t]')
        cxpb__jyhmv = ', '.join(
            'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {})[idx_t]'
            .format(df.columns.index(pzsex__zyy)) for pzsex__zyy in df.columns)
        return bodo.hiframes.dataframe_impl._gen_init_df(llc__qshk, df.
            columns, cxpb__jyhmv, jqe__vul)
    if isinstance(idx, types.BaseTuple) and len(idx) == 2:
        ouxc__ykn = idx.types[1]
        if is_overload_constant_str(ouxc__ykn):
            nwnsh__kuw = get_overload_const_str(ouxc__ykn)
            rwsi__dit = df.columns.index(nwnsh__kuw)

            def impl_col_name(I, idx):
                df = I._obj
                jqe__vul = bodo.hiframes.pd_dataframe_ext.get_dataframe_index(
                    df)
                qgn__ivz = bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df
                    , rwsi__dit)
                return bodo.hiframes.pd_series_ext.init_series(qgn__ivz,
                    jqe__vul, nwnsh__kuw).loc[idx[0]]
            return impl_col_name
        if is_overload_constant_list(ouxc__ykn):
            col_idx_list = get_overload_const_list(ouxc__ykn)
            if len(col_idx_list) > 0 and not isinstance(col_idx_list[0], (
                bool, np.bool_)) and not all(pzsex__zyy in df.columns for
                pzsex__zyy in col_idx_list):
                raise_bodo_error(
                    f'DataFrame.loc[]: invalid column list {col_idx_list}; not all in dataframe columns {df.columns}'
                    )
            return gen_df_loc_col_select_impl(df, col_idx_list)
    raise_bodo_error(
        f'DataFrame.loc[] getitem (location-based indexing) using {idx} not supported yet.'
        )


def gen_df_loc_col_select_impl(df, col_idx_list):
    if len(col_idx_list) > 0 and isinstance(col_idx_list[0], (bool, np.bool_)):
        col_idx_list = list(pd.Series(df.columns, dtype=object)[col_idx_list])
    cxpb__jyhmv = ', '.join(
        'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {})[idx[0]]'
        .format(df.columns.index(pzsex__zyy)) for pzsex__zyy in col_idx_list)
    jqe__vul = 'bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)[idx[0]]'
    llc__qshk = 'def impl(I, idx):\n'
    llc__qshk += '  df = I._obj\n'
    return bodo.hiframes.dataframe_impl._gen_init_df(llc__qshk,
        col_idx_list, cxpb__jyhmv, jqe__vul)


@overload(operator.setitem, no_unliteral=True)
def df_loc_setitem_overload(df, idx, val):
    if not isinstance(df, DataFrameLocType):
        return
    raise_bodo_error(
        f'DataFrame.loc setitem unsupported for dataframe {df.df_type}, index {idx}, value {val}'
        )


class DataFrameIatType(types.Type):

    def __init__(self, df_type):
        self.df_type = df_type
        xxv__khoxk = 'DataFrameIatType({})'.format(df_type)
        super(DataFrameIatType, self).__init__(xxv__khoxk)

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)
    ndim = 2


@register_model(DataFrameIatType)
class DataFrameIatModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        wuh__rchec = [('obj', fe_type.df_type)]
        super(DataFrameIatModel, self).__init__(dmm, fe_type, wuh__rchec)


make_attribute_wrapper(DataFrameIatType, 'obj', '_obj')


@intrinsic
def init_dataframe_iat(typingctx, obj=None):

    def codegen(context, builder, signature, args):
        fatsf__ctu, = args
        iqe__hquhu = signature.return_type
        azna__can = cgutils.create_struct_proxy(iqe__hquhu)(context, builder)
        azna__can.obj = fatsf__ctu
        context.nrt.incref(builder, signature.args[0], fatsf__ctu)
        return azna__can._getvalue()
    return DataFrameIatType(obj)(obj), codegen


@overload_attribute(DataFrameType, 'iat')
def overload_dataframe_iat(df):
    check_runtime_cols_unsupported(df, 'DataFrame.iat')
    return lambda df: bodo.hiframes.dataframe_indexing.init_dataframe_iat(df)


@overload(operator.getitem, no_unliteral=True)
def overload_iat_getitem(I, idx):
    if not isinstance(I, DataFrameIatType):
        return
    if isinstance(idx, types.BaseTuple) and len(idx) == 2:
        if not isinstance(idx.types[0], types.Integer):
            raise BodoError(
                'DataFrame.iat: iAt based indexing can only have integer indexers'
                )
        if not is_overload_constant_int(idx.types[1]):
            raise_bodo_error(
                'DataFrame.iat getitem: column index must be a constant integer. For more informaton, see https://docs.bodo.ai/latest/source/programming_with_bodo/require_constants.html.'
                )
        rwsi__dit = get_overload_const_int(idx.types[1])

        def impl_col_ind(I, idx):
            df = I._obj
            qgn__ivz = bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df,
                rwsi__dit)
            return qgn__ivz[idx[0]]
        return impl_col_ind
    raise BodoError('df.iat[] getitem using {} not supported'.format(idx))


@overload(operator.setitem, no_unliteral=True)
def overload_iat_setitem(I, idx, val):
    if not isinstance(I, DataFrameIatType):
        return
    if isinstance(idx, types.BaseTuple) and len(idx) == 2:
        if not isinstance(idx.types[0], types.Integer):
            raise BodoError(
                'DataFrame.iat: iAt based indexing can only have integer indexers'
                )
        if not is_overload_constant_int(idx.types[1]):
            raise_bodo_error(
                'DataFrame.iat setitem: column index must be a constant integer. For more informaton, see https://docs.bodo.ai/latest/source/programming_with_bodo/require_constants.html'
                )
        rwsi__dit = get_overload_const_int(idx.types[1])
        if is_immutable_array(I.df_type.data[rwsi__dit]):
            raise BodoError(
                f'DataFrame setitem not supported for column with immutable array type {I.df_type.data}'
                )

        def impl_col_ind(I, idx, val):
            df = I._obj
            qgn__ivz = bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df,
                rwsi__dit)
            qgn__ivz[idx[0]] = bodo.utils.conversion.unbox_if_timestamp(val)
        return impl_col_ind
    raise BodoError('df.iat[] setitem using {} not supported'.format(idx))


@lower_cast(DataFrameIatType, DataFrameIatType)
@lower_cast(DataFrameILocType, DataFrameILocType)
@lower_cast(DataFrameLocType, DataFrameLocType)
def cast_series_iat(context, builder, fromty, toty, val):
    azna__can = cgutils.create_struct_proxy(fromty)(context, builder, val)
    kxb__kes = context.cast(builder, azna__can.obj, fromty.df_type, toty.
        df_type)
    htp__uvsmv = cgutils.create_struct_proxy(toty)(context, builder)
    htp__uvsmv.obj = kxb__kes
    return htp__uvsmv._getvalue()
